"""
predefined params for openrtk
"""
JSON_FILE_NAME = 'openrtk.json'

def get_app_names():
    '''
    define openimu app type
    '''
    app_names = ['INS',
                 'RAWDATA',
                 'RTK',
                 ]
    return app_names


APP_STR = ['INS', 'RAWDATA', 'RTK']
